from bedrockworldoperator import Range, Dimension, ChunkPos
from bedrockworldoperator import RANGE_OVERWORLD, RANGE_NETHER, RANGE_END
from bedrockworldoperator import DIMENSION_OVERWORLD, DIMENSION_NETHER, DIMENSION_END

from .timeline.timeline_database import new_timeline_database



import ctypes
import platform
import os
import sys
from pathlib import Path


package_dir = Path(__file__).parent.parent.resolve()
lib_path = os.path.join(package_dir, "dynamic_libs")

system = platform.system().lower()
arch = platform.machine().lower()

format_arch = ""
if sys.maxsize == 2**63 - 1:
    if arch in ("x86_64", "amd64"):
        format_arch = "amd64"
    elif arch in ("arm64", "aarch64"):
        format_arch = "arm64"
else:
    if arch in ("x86", "i686"):
        format_arch = "x86"


LIB = None

match system:
    case "windows":
        if format_arch == "amd64":
            LIB = ctypes.cdll.LoadLibrary(
                os.path.join(lib_path, "bedrock-chunk-diff_windows_amd64.dll")
            )
        elif format_arch == "x86":
            LIB = ctypes.cdll.LoadLibrary(
                os.path.join(lib_path, "bedrock-chunk-diff_windows_x86.dll")
            )
    case "darwin":
        if format_arch == "amd64":
            LIB = ctypes.cdll.LoadLibrary(
                os.path.join(lib_path, "bedrock-chunk-diff_macos_amd64.dylib")
            )
        elif format_arch == "arm64":
            LIB = ctypes.cdll.LoadLibrary(
                os.path.join(lib_path, "bedrock-chunk-diff_macos_arm64.dylib")
            )
    case _:
        if format_arch == "amd64":
            LIB = ctypes.cdll.LoadLibrary(
                os.path.join(lib_path, "bedrock-chunk-diff_linux_amd64.so")
            )
        elif format_arch == "arm64":
            if arch == "aarch64":
                LIB = ctypes.cdll.LoadLibrary(
                    os.path.join(lib_path, "bedrock-chunk-diff_android_arm64.so")
                )
            else:
                LIB = ctypes.cdll.LoadLibrary(
                    os.path.join(lib_path, "bedrock-chunk-diff_linux_arm64.so")
                )

if LIB is None:
    raise Exception(f"Your machine (system={system}, arch={arch}) is not supported.")



import numpy
from .types import LIB
from .types import CInt, CLongLong, CString, CSlice
from .types import as_c_bytes, as_python_bytes, as_python_string
from .utils import pack_bytes_list, unpack_next_or_last


LIB.AppendDiskChunk.argtypes = [CLongLong, CSlice, CSlice, CInt, CInt]
LIB.AppendNetworkChunk.argtypes = [CLongLong, CSlice, CSlice, CInt, CInt]
LIB.Empty.argtypes = [CLongLong]
LIB.ReadOnly.argtypes = [CLongLong]
LIB.AllTimePoint.argtypes = [CLongLong]
LIB.AllTimePointLen.argtypes = [CLongLong]
LIB.SetMaxLimit.argtypes = [CLongLong, CInt]
LIB.Compact.argtypes = [CLongLong]
LIB.NextDiskChunk.argtypes = [CLongLong]
LIB.NextNetworkChunk.argtypes = [CLongLong]
LIB.LastDiskChunk.argtypes = [CLongLong]
LIB.LastNetworkChunk.argtypes = [CLongLong]
LIB.Pop.argtypes = [CLongLong]
LIB.Save.argtypes = [CLongLong]

LIB.AppendDiskChunk.restype = CString
LIB.AppendNetworkChunk.restype = CString
LIB.Empty.restype = CInt
LIB.ReadOnly.restype = CInt
LIB.AllTimePoint.restype = CSlice
LIB.AllTimePointLen.restype = CInt
LIB.SetMaxLimit.restype = CString
LIB.Compact.restype = CString
LIB.NextDiskChunk.restype = CSlice
LIB.NextNetworkChunk.restype = CSlice
LIB.LastDiskChunk.restype = CSlice
LIB.LastNetworkChunk.restype = CSlice
LIB.Pop.restype = CString
LIB.Save.restype = CString


def ctl_append_disk_chunk(
    id: int,
    chunk_payload: list[bytes],
    nbt_payload: list[bytes],
    range_start: int,
    range_end: int,
) -> str:
    return as_python_string(
        LIB.AppendDiskChunk(
            CLongLong(id),
            as_c_bytes(pack_bytes_list(chunk_payload)),
            as_c_bytes(pack_bytes_list(nbt_payload)),
            CInt(range_start),
            CInt(range_end),
        )
    )


def ctl_append_network_chunk(
    id: int,
    chunk_payload: list[bytes],
    nbt_payload: list[bytes],
    range_start: int,
    range_end: int,
) -> str:
    return as_python_string(
        LIB.AppendNetworkChunk(
            CLongLong(id),
            as_c_bytes(pack_bytes_list(chunk_payload)),
            as_c_bytes(pack_bytes_list(nbt_payload)),
            CInt(range_start),
            CInt(range_end),
        )
    )


def ctl_empty(id: int) -> int:
    return int(LIB.Empty(CLongLong(id)))


def ctl_read_only(id: int) -> int:
    return int(LIB.ReadOnly(CLongLong(id)))


def ctl_all_time_point(id: int) -> numpy.ndarray:
    return numpy.frombuffer(
        as_python_bytes(LIB.AllTimePoint(CLongLong(id))), dtype="<i8"
    )


def ctl_all_time_point_len(id: int) -> int:
    return int(LIB.AllTimePointLen(CLongLong(id)))


def ctl_set_max_limit(id: int, max_limit: int) -> str:
    return as_python_string(LIB.SetMaxLimit(CLongLong(id), CInt(max_limit)))


def ctl_compact(id: int) -> str:
    return as_python_string(LIB.Compact(CLongLong(id)))


def ctl_next_disk_chunk(
    id: int,
) -> tuple[list[bytes], int, int, list[bytes], int, bool]:
    return unpack_next_or_last(as_python_bytes(LIB.NextDiskChunk(CLongLong(id))), True)


def ctl_next_network_chunk(
    id: int,
) -> tuple[list[bytes], int, int, list[bytes], int, bool]:
    return unpack_next_or_last(
        as_python_bytes(LIB.NextNetworkChunk(CLongLong(id))), True
    )


def ctl_last_disk_chunk(id: int) -> tuple[list[bytes], int, int, list[bytes], int]:
    sub_chunks, range_start, range_end, nbts, update_unix_time, _ = unpack_next_or_last(
        as_python_bytes(LIB.LastDiskChunk(CLongLong(id))), False
    )
    return sub_chunks, range_start, range_end, nbts, update_unix_time


def ctl_last_network_chunk(id: int) -> tuple[list[bytes], int, int, list[bytes], int]:
    sub_chunks, range_start, range_end, nbts, update_unix_time, _ = unpack_next_or_last(
        as_python_bytes(LIB.LastNetworkChunk(CLongLong(id))), False
    )
    return sub_chunks, range_start, range_end, nbts, update_unix_time


def ctl_pop(id: int) -> str:
    return as_python_string(LIB.Pop(CLongLong(id)))


def ctl_save(id: int) -> str:
    return as_python_string(LIB.Save(CLongLong(id)))



import struct
from io import BytesIO


def unpack_bytes_list(payload: bytes) -> list[bytes]:
    result: list[bytes] = []
    ptr = 0

    while ptr < len(payload):
        length: int = struct.unpack("<I", payload[ptr : ptr + 4])[0]
        result.append(payload[ptr + 4 : ptr + 4 + length])
        ptr = ptr + 4 + length

    return result


def pack_bytes_list(sub_chunks: list[bytes]) -> bytes:
    w = BytesIO()
    for i in sub_chunks:
        w.write(struct.pack("<I", len(i)))
        w.write(i)
    return w.getvalue()


def unpack_next_or_last(
    payload: bytes, read_is_last_element: bool
) -> tuple[list[bytes], int, int, list[bytes], int, bool]:
    r = BytesIO(payload)

    length: int = struct.unpack("<I", r.read(4))[0]
    chunk_payload_bytes = r.read(length)
    sub_chunks = unpack_bytes_list(chunk_payload_bytes)

    range_start: int = struct.unpack("<h", r.read(2))[0]
    range_end: int = struct.unpack("<h", r.read(2))[0]

    length = struct.unpack("<I", r.read(4))[0]
    nbt_payload = r.read(length)
    nbts = unpack_bytes_list(nbt_payload)

    update_unix_time: int = struct.unpack("<q", r.read(8))[0]

    if read_is_last_element:
        is_last_element = bool(int(r.read(1)))
        return (
            sub_chunks,
            range_start,
            range_end,
            nbts,
            update_unix_time,
            is_last_element,
        )
    else:
        return sub_chunks, range_start, range_end, nbts, update_unix_time, False
